import datetime
import re
from pathlib import Path
from typing import List, Optional, Union, Tuple
from terracatalogueclient import Catalogue, Product
from terracatalogueclient.config import CatalogueConfig, CatalogueEnvironment
from datetime import date
from dateutil.relativedelta import relativedelta
import os

MAX_OSCARS = 10e3
MAX_YEARS = 35
CHANGING_HREFS = ["clms_global_wl_lakes_v2_daily_geojson", "clms_global_wl_rivers_v2_daily_geojson"]
SWI_SPLIT_MANIFEST = ["clms_global_swi_12.5km_v3_static_netcdf", "clms_global_swi_12.5km_v3_static_netcdf"]
LSP_SPLIT_MANIFEST = ["clms_global_lsp_300m_v1_yearly_geotiff"]

def generate_all_manifests(manifest_root_path: str,
                           index_list_path: Optional = None,
                           publication_date: Optional[Tuple[None, str]] = None,
                           dry_run: bool = False) -> None:
    """
    Function that generates all manifest files and subdirectories in the manifest_root_path directory. The index_list_path
    and publication_date parameters are optional and mainly used for testing.
    """

    # Start a CGSLOPS catalogue.
    catalogue = Catalogue(CatalogueConfig.from_environment(CatalogueEnvironment.CGLS))
    # Retrieve all indices from the catalogue, generate a manifest for each one.
    for product_index in get_all_product_indices(catalogue):
        # Check if a manifest file is needed. Otherwise skip the index.
        if should_make_manifest(product_index, index_list_path):
            make_and_save_manifests(manifest_root_path, catalogue, product_index, publication_date, dry_run)


def _per_productGroupId(manifest_root_path, catalogue, product_index, publication_date):
    collection_info = [collection for collection in catalogue.get_collections() if collection.id == product_index][0]
    for productGroupId in collection_info.properties["productInformation"]["productGroupId"]:
        query = f'properties.productInformation.productGroupId:"{productGroupId}"'
        products = catalogue.get_products(collection_info.id, publicationDate=publication_date, q=query)
        write_results(products, get_manifest_path(manifest_root_path, product_index, productGroupId))


def make_and_save_manifests(manifest_root_path, catalogue, product_index, publication_date, dry_run=False):
    if "lsp_300m_v1_yearly_geotiff" in product_index:
        _per_product_type_and_productGroupId(manifest_root_path, catalogue, product_index, publication_date)
    elif "swi_12.5km_v1_static_netcdf" in product_index or "swi_12.5km_v3_static_netcdf" in product_index:
        _per_productGroupId(manifest_root_path, catalogue, product_index, publication_date)
    else:
        # Generate the content of the manifest files.
        if not dry_run:
            products: List[str] = generate_manifest(catalogue, product_index, publication_date=publication_date)
        else:
            products = ["dummy"]
        # Write the content of the manifest file in the right (sub)directory.
        write_results(products, get_manifest_path(manifest_root_path, product_index))


def _per_product_type_and_productGroupId(manifest_root_path, catalogue, product_index, publication_date):
    collection_info = [collection for collection in catalogue.get_collections() if collection.id == product_index][0]
    for productGroupId in collection_info.properties["productInformation"]["productGroupId"]:
        for productType in collection_info.properties["productInformation"]["productType"]:
            query = (f'properties.productInformation.productGroupId:"{productGroupId}" '
                     f'AND properties.productInformation.productType:"{productType}"')
            products = catalogue.get_products(collection_info.id, publicationDate=publication_date, q=query)
            write_results(products, get_manifest_path(manifest_root_path, product_index, productGroupId, productType))


def get_all_product_indices(catalogue: Catalogue) -> List[str]:
    return [collection.id for collection in catalogue.get_collections()]


def should_make_manifest(collection_id: str, index_list_path: Optional[str]) -> bool:
    # If an index_list_path is provided, only generate a manifest if the collection_id is in the list.
    if index_list_path is not None:
        with open(index_list_path, 'r') as f:
            indices = f.read().splitlines()
            return collection_id in indices
    else:
        allowed_indices_geotiff = [
            "clms_global_npp_300m_v1_10daily_geotiff",
            "clms_global_ba_300m_v3_daily_geotiff",
            "clms_global_gpp_300m_v1_10daily_geotiff",
            "clms_global_ba_300m_v3_monthly_geotiff"
        ]
        return ("netcdf" in collection_id or
                "lsp" in collection_id or
                collection_id in allowed_indices_geotiff or
                "geojson" in collection_id)


def generate_manifest(catalogue: Catalogue, index: str, publication_date=None) -> List[str]:
    return [extract_href(product) for product in retrieve_all_products(catalogue, index, publication_date)]


def extract_href(product: Product) -> str:
    return product.properties["links"]["data"][0]["href"]


def retrieve_all_products(catalogue: Catalogue, index: str, publication_date=None) -> List[Product]:
    total_amount_products = catalogue.get_product_count(index, publicationDate=publication_date)

    # Check if the amount of products doesn't exceed the maximum for oscars. If it does, split the retrieval in years.
    if total_amount_products < MAX_OSCARS:
        products = [product for product in catalogue.get_products(index, publicationDate=publication_date)]
    else:
        products = [product for product in retrieve_all_products_per_year(catalogue, index, publication_date)]

    # Check if the amount of products is correct, otherwise raise an error.
    check_product_amount(total_amount_products, products, index=index)
    return products


def retrieve_all_products_per_year(catalogue: Catalogue, index: str, publication_date=None) -> List[Product]:
    products = []
    for years_in_past in range(MAX_YEARS):
        start_of_year = date.today() - relativedelta(years=years_in_past + 1)
        end_of_year = date.today() - relativedelta(years=years_in_past)
        new_products = retrieve_for_year(catalogue, index, start_of_year, end_of_year,
                                         publication_date=publication_date)
        products.extend(new_products)
    return products


def retrieve_for_year(catalogue, index, start_of_year, end_of_year, publication_date=None):
    expected_amount_products = catalogue.get_product_count(index, publicationDate=publication_date,
                                                           q=get_query(start_of_year, end_of_year))
    # Check if the amount of products doesn't exceed the maximum for oscars. If it does, split the retrieval in months.
    if expected_amount_products < MAX_OSCARS:
        new_products = [product for product in catalogue.get_products(index, publicationDate=publication_date,
                                                                      q=get_query(start_of_year, end_of_year))]
    else:
        new_products = get_products_per_month(catalogue, index, start_of_year, end_of_year,
                                              publication_date=publication_date)

    # Check if the amount of products is correct, otherwise raise an error.
    check_product_amount(expected_amount_products, new_products, index=index)
    return new_products


def get_query(begin_date, end_date):
    time_field = "properties.acquisitionInformation.acquisitionParameters.beginningDateTime"
    begin_date_zulu = f"{begin_date}T00:00:00.000Z"
    actual_end_date = end_date if end_date == date.today() else end_date - relativedelta(days=1)
    end_date_zulu = f"{actual_end_date}T23:59:59.999Z"
    return f"{time_field}:[{begin_date_zulu} TO {end_date_zulu}]"


def get_products_per_month(catalogue: Catalogue, index: str, start_date: datetime.date, end_date: datetime.date,
                           publication_date=None):
    all_products = []
    begin_current_date = start_date
    for _ in range(12):
        end_current_month = begin_current_date + relativedelta(months=1)
        new_products = [product for product in catalogue.get_products(index, publicationDate=publication_date,
                                                                      q=get_query(begin_current_date,
                                                                                  min(end_current_month, end_date)))]
        all_products = all_products + new_products
        begin_current_date = end_current_month

        if end_current_month > end_date:
            # No products after end_date should be retrieved, stop the loop if the begin and end date are not exactly one
            # year apart.
            break

    return all_products


def check_product_amount(total_amount_products, products, index=None) -> None:
    if total_amount_products == len(products):
        return
    else:
        if total_amount_products > len(products):
            raise Exception(
                f"Missing products. Expected {total_amount_products}, got {len(products)} for index {index}.")
        else:
            raise Exception(
                f"too many products. Expected {total_amount_products}, got {len(products)} for index {index}.")


def remove_geotiff_if_needed(name: str) -> str:
    # All geotiffs (except lsp!) need to have the geotiff suffix removed.
    should_remove = (not "lsp" in name) and "geotiff" in name
    return name.replace("_geotiff", "") if should_remove else name


def replace_prefix_if_needed(name: str) -> str:
    # All geotiffs need the cgls-prefix.
    should_add_prefix = "geotiff" in name
    return name.replace('clms_global', 'cgls') if should_add_prefix else name


def remove_prefix_if_needed(product_index: str, not_remove=False) -> str:
    # All prefixes of subdirectories need to be removed, also all wl_rivers products don't need a prefix.
    return product_index.replace("clms_global_", "") if not not_remove else product_index


def get_subdirectory_name(product_index: str) -> str:
    return remove_geotiff_if_needed(remove_prefix_if_needed(product_index))


def get_manifest_path(manifest_root_path, product_index, productGroupId=None, productType=None):
    return Path(
        manifest_root_path) / f"{get_subdirectory_name(product_index)}/{get_manifest_name(product_index, productGroupId, productType)}"


def get_manifest_name(product_index: str, productGroupId=None, productType=None) -> str:
    manifest_name = f"manifest_{product_index}_latest.txt"
    is_edge_case = productGroupId or productType

    if not is_edge_case:
        manifest_name = replace_prefix_if_needed(manifest_name)
        manifest_name = remove_prefix_if_needed(manifest_name, not_remove=("wl_rivers" not in product_index))
        manifest_name = remove_geotiff_if_needed(manifest_name)
        return manifest_name

    # Edge case 1: lsp collection is split over multiple manifests.
    lsp_substring = "clms_global_lsp"
    is_lsp = lsp_substring in product_index
    if is_lsp:
        split_parts = manifest_name.split(lsp_substring)
        extra_info = f"-{productType.lower()}-{productGroupId.lower()}"
        return split_parts[0] + lsp_substring + extra_info + split_parts[1]

    # Edge case 2: some swi collections are split into multiple manifests.
    swi_substring = "clms_global_swi"
    is_split_swi = "swi_12.5km_v1_static_netcdf" in product_index or "swi_12.5km_v3_static_netcdf" in product_index
    if is_split_swi:
        split_parts = manifest_name.split(swi_substring)
        extra_info = f"-{productGroupId.lower()}"
        return split_parts[0] + swi_substring + extra_info + split_parts[1]


def write_results(results: List[str], filename: Union[str, Path]) -> None:
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w+') as f:
        for href in results:
            f.write("%s\n" % href)


def are_equal_hrefs(other_href, product_href):
    parent_product, filename_product = os.path.split(product_href)
    parent_other, filename_other = os.path.split(other_href)

    pattern = r'(\d+_[A-Z]+_V\d+\.\d+\.\d+)'
    id_product = re.search(pattern, filename_product).group(1)
    id_other = re.search(pattern, filename_other).group(1)

    return parent_product == parent_other and id_product == id_other


def add_to_manifest(manifest_root_path: str, index: str, product_id: str) -> None:
    """
    Function that adds a product to the manifest file of a given index.

    :param manifest_root_path: the root path of the manifest files
    :param index: the index of the product
    :param product_id: the id of the product
    """
    # Only add to manifest if the index has a manifest.
    if should_make_manifest(index, None):
        manifest_path, current_products, product_href = _get_details(manifest_root_path, index, product_id)

        if index not in CHANGING_HREFS:
            current_products.append(product_href)
            current_products = list(set(current_products))  # Remove duplicates
            write_results(current_products, manifest_path)
        else:
            resulting_products = [product_href]
            for other_href in current_products:
                if not are_equal_hrefs(other_href, product_href):
                    resulting_products.append(other_href)

            write_results(resulting_products, manifest_path)

def remove_from_manifest(manifest_root_path: str, index: str, product_id: str) -> None:
    """
    Function that removes a product to the manifest file of a given index.

    :param manifest_root_path: the root path of the manifest files
    :param index: the index of the product
    :param product_id: the id of the product
    """
    # Only remove from manifest if the index has a manifest.
    if should_make_manifest(index, None):
        manifest_path, current_products, product_href = _get_details(manifest_root_path, index, product_id)
        if index not in CHANGING_HREFS:
            if product_href in current_products:
                current_products.remove(product_href)
            write_results(current_products, manifest_path)
        else:
            resulting_products = []
            for other_href in current_products:
                if not are_equal_hrefs(other_href, product_href):
                    resulting_products.append(other_href)
            write_results(resulting_products, manifest_path)

def _get_details(manifest_root_path: Union[Path, str], index: str, product_id: str) -> Tuple[Path, List[str], str]:
    """
    Function that retrieves the path to the manifest file, the current products in the manifest and the href of the product.
    """
    catalogue = Catalogue(CatalogueConfig.from_environment(CatalogueEnvironment.CGLS))
    product = list(catalogue.get_products(index, q=f"properties.identifier:{product_id}"))[0]
    product_href = product.properties["links"]["data"][0]["href"]

    product_type, product_group = None, None
    if index in LSP_SPLIT_MANIFEST:
        product_type = product.properties["productInformation"]["productType"]
        product_group = product.properties["productInformation"]["productGroupId"]
    elif index in SWI_SPLIT_MANIFEST:
        product_type = product.properties["productInformation"]["productType"]
        product_group = product.properties["productInformation"]["productGroupId"]

    manifest_path = get_manifest_path(manifest_root_path, index, productGroupId=product_group, productType=product_type)

    with open(manifest_path, 'r') as f:
        current_products = f.read().splitlines()

    return manifest_path, current_products, product_href
